/*
 * Copyright 1999-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 */

package sun.java2d.xr;

import java.awt.*;
import java.awt.MultipleGradientPaint.*;
import java.awt.Transparency;
import java.awt.geom.*;
import java.awt.image.*;
import sun.awt.*;
import sun.font.*;
import sun.java2d.InvalidPipeException;
import sun.java2d.SunGraphics2D;
import sun.java2d.SurfaceData;
import sun.java2d.SurfaceDataProxy;
import sun.java2d.loops.*;
import sun.java2d.pipe.*;
import sun.java2d.x11.*;
import static sun.java2d.xr.XRUtils.XDoubleToFixed;

public abstract class XRSurfaceData extends XSurfaceData {
    X11ComponentPeer peer;
    XRGraphicsConfig graphicsConfig;
    private RenderLoops solidloops;

    protected int depth;

    private static native void initIDs(boolean gradCache);

    protected native void XRInitSurface(int depth, int width, int height, long drawable, int pictFormat);

    native void XRInitXRender(long xsdo, int pictForm);

    public static final String DESC_BYTE_A8_X11 = "Byte A8 Pixmap";
    public static final String DESC_INT_RGB_X11 = "Integer RGB Pixmap";
    public static final String DESC_INT_ARGB_X11 = "Integer ARGB-Pre Pixmap";

    public static final SurfaceType ByteA8X11 = SurfaceType.ByteGray.deriveSubType(DESC_BYTE_A8_X11);
    public static final SurfaceType IntRgbX11 = SurfaceType.IntRgb.deriveSubType(DESC_INT_RGB_X11);
    public static final SurfaceType IntArgbX11 = SurfaceType.IntArgbPre.deriveSubType(DESC_INT_ARGB_X11);

    public Raster getRaster(int x, int y, int w, int h) {
	throw new InternalError("not implemented yet");
    }

    protected XRRenderer xrpipe;
    protected PixelToShapeConverter xrtxpipe;
    protected TextPipe xrtextpipe;
    protected XRDrawImage xrDrawImage;

    public static void initXRSurfaceData() {
	if (!X11GraphicsEnvironment.isX11SurfaceDataInitialized()) {
	    boolean cacheGradients = true;
	    String xProp = System.getProperty("sun.java2d.xrgradcache");
	    if (xProp != null) {
		cacheGradients = !(xProp.equalsIgnoreCase("false") || xProp.equalsIgnoreCase("f"));
	    }
	    initIDs(cacheGradients);

	    XRPMBlitLoops.register();
	    XRMaskFill.register();
	    XRMaskBlit.register();

	    X11GraphicsEnvironment.setX11SurfaceDataInitialized();
	}
    }

    /**
     * Synchronized accessor method for isDrawableValid.
     */
    protected boolean isXRDrawableValid() {
	try {
	    SunToolkit.awtLock();
	    return isDrawableValid();
	} finally {
	    SunToolkit.awtUnlock();
	}
    }

    @Override
    public SurfaceDataProxy makeProxyFor(SurfaceData srcData) {
	return XRSurfaceDataProxy.createProxy(srcData, graphicsConfig);
    }

    public void validatePipe(SunGraphics2D sg2d) {
	TextPipe textpipe;
	boolean validated = false;

	/*
	 * The textpipe for now can't handle TexturePaint when extra-alpha is
	 * specified nore XOR mode
	 */
	if (sg2d.compositeState < SunGraphics2D.COMP_XOR
		&& (sg2d.paintState < SunGraphics2D.PAINT_TEXTURE || sg2d.composite == null || !(sg2d.composite instanceof AlphaComposite) || ((AlphaComposite) sg2d.composite)
			.getAlpha() == 1.0f)) {
	    textpipe = xrtextpipe;
	} else {
	    super.validatePipe(sg2d);
	    textpipe = sg2d.textpipe;
	    validated = true;
	}

	PixelToShapeConverter txPipe = null;
	XRRenderer nonTxPipe = null;

	if (sg2d.antialiasHint != SunHints.INTVAL_ANTIALIAS_ON) {
	    if (sg2d.paintState <= SunGraphics2D.PAINT_ALPHACOLOR) {
		if (sg2d.compositeState < SunGraphics2D.COMP_XOR) {
		    txPipe = xrtxpipe;
		    nonTxPipe = xrpipe;
		}
	    } else if (sg2d.compositeState <= SunGraphics2D.COMP_ALPHA) {
		if (isPaintValid(sg2d)) {
		    txPipe = xrtxpipe;
		    nonTxPipe = xrpipe;
		}
		// custom paints handled by super.validatePipe() below
	    }
	}

	if (txPipe != null) {
	    if (sg2d.transformState >= SunGraphics2D.TRANSFORM_TRANSLATESCALE) {
		sg2d.drawpipe = txPipe;
		sg2d.fillpipe = txPipe;
	    } else if (sg2d.strokeState != SunGraphics2D.STROKE_THIN) {
		sg2d.drawpipe = txPipe;
		sg2d.fillpipe = nonTxPipe;
	    } else {
		sg2d.drawpipe = nonTxPipe;
		sg2d.fillpipe = nonTxPipe;
	    }
	    sg2d.shapepipe = nonTxPipe;
	} else {
	    if (!validated) {
		super.validatePipe(sg2d);
	    }
	}

	// install the text pipe based on our earlier decision
	sg2d.textpipe = textpipe;

	// always override the image pipe with the specialized OGL pipe
	sg2d.imagepipe = xrDrawImage;
    }

    /**
     * Returns an accalerated MaskFill object if the current paint is supported
     * by the pipeline.
     */
    protected MaskFill getMaskFill(SunGraphics2D sg2d) {
	if (sg2d.paintState > SunGraphics2D.PAINT_ALPHACOLOR && !isPaintValid(sg2d)) {
	    return null;
	}
	return super.getMaskFill(sg2d);
    }

    public RenderLoops getRenderLoops(SunGraphics2D sg2d) {
	if (sg2d.paintState <= SunGraphics2D.PAINT_ALPHACOLOR && sg2d.compositeState <= SunGraphics2D.COMP_ALPHA) {
	    return solidloops;
	}

	return super.getRenderLoops(sg2d);
    }

    public GraphicsConfiguration getDeviceConfiguration() {
	return graphicsConfig;
    }

    /**
     * Method for instantiating a Window SurfaceData
     */
    public static XRWindowSurfaceData createData(X11ComponentPeer peer) {
	XRGraphicsConfig gc = getGC(peer);
	return new XRWindowSurfaceData(peer, gc, gc.getSurfaceType());
    }

    /**
     * Method for instantiating a Pixmap SurfaceData (offscreen). If the surface
     * is opaque a 24-bit/RGB surface is chosen, otherwise a 32-bit ARGB
     * surface.
     */
    public static XRPixmapSurfaceData createData(XRGraphicsConfig gc, int width, int height, ColorModel cm, Image image, long drawable,
	    int transparency) {
	int depth = transparency > Transparency.OPAQUE ? 32 : 24;
	if (depth == 24) {
	    cm = new DirectColorModel(depth, 0x00FF0000, 0x0000FF00, 0x000000FF);
	} else {
	    cm = new DirectColorModel(depth, 0x00FF0000, 0x0000FF00, 0x000000FF, 0xFF000000);
	}

	return new XRPixmapSurfaceData(gc, width, height, image, getSurfaceType(gc, transparency), cm, drawable, transparency, XRUtils
		.getPictureFormatForTransparency(transparency), depth);
    }

    protected XRSurfaceData(X11ComponentPeer peer, XRGraphicsConfig gc, SurfaceType sType, ColorModel cm, int depth, int transparency) {
	super(sType, cm);
	this.peer = peer;
	this.graphicsConfig = gc;
	this.solidloops = graphicsConfig.getSolidLoops(sType);
	this.depth = depth;
	initOps(peer, graphicsConfig, depth);

	setBlitProxyKey(gc.getProxyKey());
    }

    /**
     * Inits the XRender-data-structures which belong to the XRSurfaceData.
     * 
     * @param pictureFormat
     */
    public void initXRender(int pictureFormat) {
	try {
	    SunToolkit.awtLock();
	    XRInitXRender(getNativeOps(), pictureFormat);
	} finally {
	    SunToolkit.awtUnlock();
	}
    }

    public static XRGraphicsConfig getGC(X11ComponentPeer peer) {
	if (peer != null) {
	    return (XRGraphicsConfig) peer.getGraphicsConfiguration();
	} else {
	    GraphicsEnvironment env = GraphicsEnvironment.getLocalGraphicsEnvironment();
	    GraphicsDevice gd = env.getDefaultScreenDevice();
	    return (XRGraphicsConfig) gd.getDefaultConfiguration();
	}
    }

    /**
     * Returns a boolean indicating whether or not a copyArea from the given
     * rectangle source coordinates might be incomplete and result in X11
     * GraphicsExposure events being generated from XCopyArea. This method
     * allows the SurfaceData copyArea method to determine if it needs to set
     * the GraphicsExposures attribute of the X11 GC to True or False to receive
     * or avoid the events.
     * 
     * @return true if there is any chance that an XCopyArea from the given
     *         source coordinates could produce any X11 Exposure events.
     */
    public abstract boolean canSourceSendExposures(int x, int y, int w, int h);

    /**
     * CopyArea is implemented using the "old" X11 GC, therefor clip and
     * needExposures have to be validated against that GC. Pictures and GCs
     * don't share state.
     */
    public void validateCopyAreaGC(Region gcClip, boolean needExposures) {
	if (validatedGCClip != gcClip) {
	    if (gcClip != null) {
		XSetClip(xgc, gcClip.getLoX(), gcClip.getLoY(), gcClip.getHiX(), gcClip.getHiY(), gcClip.isRectangular() ? null : gcClip);
	    } else {
		XResetClip(xgc);
	    }
	    validatedGCClip = gcClip;
	}

	if (validatedExposures != needExposures) {
	    validatedExposures = needExposures;
	    XSetGraphicsExposures(xgc, needExposures);
	}
    }

    public boolean copyArea(SunGraphics2D sg2d, int x, int y, int w, int h, int dx, int dy) {
	if (xrpipe == null) {
	    if (!isXRDrawableValid()) {
		return true;
	    }
	    makePipes();
	}
	CompositeType comptype = sg2d.imageComp;
	if (sg2d.transformState < SunGraphics2D.TRANSFORM_TRANSLATESCALE
		&& (CompositeType.SrcOverNoEa.equals(comptype) || CompositeType.SrcNoEa.equals(comptype))) {
	    x += sg2d.transX;
	    y += sg2d.transY;
	    try {
		SunToolkit.awtLock();
		boolean needExposures = canSourceSendExposures(x, y, w, h);
		validateCopyAreaGC(sg2d.getCompClip(), needExposures);
		xrpipe.devCopyArea(getNativeOps(), xgc, x, y, x + dx, y + dy, w, h);
	    } finally {
		SunToolkit.awtUnlock();
	    }
	    return true;
	}
	return false;
    }

    /**
     * Returns the XRender SurfaceType which is able to fullfill the specified
     * transparency requirement.
     */
    public static SurfaceType getSurfaceType(XRGraphicsConfig gc, int transparency) {
	SurfaceType sType = null;

	switch (transparency) {
	case Transparency.OPAQUE:
	    sType = XRSurfaceData.IntRgbX11;
	    break;

	case Transparency.BITMASK:
	case Transparency.TRANSLUCENT:
	    sType = XRSurfaceData.IntArgbX11;
	    break;
	}

	return sType;
    }

    public void invalidate() {
	if (isValid()) {
	    setInvalid();
	    super.invalidate();
	}
    }

    /**
     * Native methods used to set the state of the native XRender surface
     */
    private static native void XRSetForeground(int pixel);

    private static native void XRSetComposite(int compRule, float eAlpha);

    private static native void XRSetClip(long xsdo, int lox, int loy, int hix, int hiy, Region complexclip);

    private static native void XRResetClip(long xsdo);

    private static native void XRSetTransformNative(long xsdo, int m00, int m01, int m02, int m10, int m11, int m12);

    private static native void XRSetTexturePaint(long xsdo);

    private static native void XRResetPaint();

    private static native void XRSetRepeat(long xsdo, int repeat);

    private static native void XRSetFilter(long xsdo, int value);

    private static native void XRSetLinearGradientPaint(float[] fractions, int[] pixels, int x1, int y1, int x2, int y2, int numStops,
	    int cycleMethod, boolean useMask, boolean linear, int m00, int m01, int m02, int m10, int m11, int m12);

    private native static void XRSetRadialGradientPaint(float[] fractions, int[] pixels, int fx, int numStops, int cycleMethod, boolean useMask,
	    boolean linear, int innerRadius, int outerRadius, int m00, int m01, int m02, int m10, int m11, int m12);

    private void XRSetTransform(AffineTransform transform) {
	double[] transformData = new double[6];
	transform.getMatrix(transformData);

	XRSetTransformNative(getNativeOps(), XDoubleToFixed(transformData[0]), XDoubleToFixed(transformData[2]), XDoubleToFixed(transformData[4]),
		XDoubleToFixed(transformData[1]), XDoubleToFixed(transformData[3]), XDoubleToFixed(transformData[5]));
    }

    private long xgc; // GC is still used for copyArea

    public static int validatedPixel = -1;
    public static int validatedPaintState = -1;
    private static Composite validatedComp;
    public static Paint validatedPaint;

    private Region validatedClip;
    private Region validatedGCClip;
    private boolean validatedExposures = true;
    private float validatedExtraAlpha = 1.0f;
    private int validatedFlags;

    private boolean transformInUse = false;
    private AffineTransform validatedSourceTransform = new AffineTransform();
    private int validatedRepeat = XRUtils.RepeatNone;
    private int validatedInterpolation = -1;
    /*
     * Used if API does not allow to pass interpolation value at the level
     * needed
     */
    private int preferredInterpolation = -1;

    public void setPreferredInterpolation(int interpolation) {
	this.preferredInterpolation = interpolation;
    }

    /**
     * Validate the source with the preferred interpolation set sometimes
     * earlier.
     * 
     * @param sxForm
     * @param repeat
     */
    void validateAsSource(AffineTransform sxForm, int repeat) {
	validateAsSource(sxForm, repeat, preferredInterpolation);
    }

    /**
     * Validates an XRSurfaceData when used as source. Note that the clip is
     * applied when used as source as well as destination.
     */
    void validateAsSource(AffineTransform sxForm, int repeat, int interpolation) {

	if (validatedClip != null) {
	    validatedClip = null;
	    XRResetClip(getNativeOps());
	}

	if (validatedRepeat != repeat) {
	    validatedRepeat = repeat;
	    XRSetRepeat(getNativeOps(), repeat);
	}

	if (sxForm == null) {
	    if (transformInUse) {
		validatedSourceTransform.setToIdentity();
		XRSetTransform(validatedSourceTransform);
		transformInUse = false;
	    }
	} else {
	    if (!transformInUse || (transformInUse && !sxForm.equals(validatedSourceTransform))) {

		validatedSourceTransform.setTransform(sxForm.getScaleX(), sxForm.getShearY(), sxForm.getShearX(), sxForm.getScaleY(), sxForm
			.getTranslateX(), sxForm.getTranslateY());

		XRSetTransform(validatedSourceTransform);
		transformInUse = true;
	    }
	}

	if (interpolation != validatedInterpolation) {
	    XRSetFilter(getNativeOps(), interpolation);
	    validatedInterpolation = interpolation;
	}
    }

    /**
     * Utility method for setting the Composite, passing down the arguments to
     * the native method.
     * 
     * @param comp
     */
    private void setComposite(Composite comp) {
	if (comp instanceof AlphaComposite) {
	    AlphaComposite aComp = (AlphaComposite) comp;
	    validatedExtraAlpha = aComp.getAlpha();

	    XRSetComposite(XRUtils.j2dAlphaCompToXR(aComp.getRule()), validatedExtraAlpha);
	} else {
	    throw new InternalError("Composite accaleration not implemented for: " + comp.getClass().getName());
	}
    }

    /**
     * @see: BufferedPaints
     */
    public int[] convertToIntArgbPixels(Color[] colors, boolean linear) {
	int[] pixels = new int[colors.length];
	for (int i = 0; i < colors.length; i++) {
	    pixels[i] = colorToIntArgbPixel(colors[i], linear);
	}
	return pixels;
    }

    /**
     * @see: BufferedPaints
     */
    public int colorToIntArgbPixel(Color c, boolean linear) {
	int rgb = c.getRGB();

	int a = rgb >>> 24;
	int r = (rgb >> 16) & 0xff;
	int g = (rgb >> 8) & 0xff;
	int b = (rgb) & 0xff;
	if (linear) {
	    r = BufferedPaints.convertSRGBtoLinearRGB(r);
	    g = BufferedPaints.convertSRGBtoLinearRGB(g);
	    b = BufferedPaints.convertSRGBtoLinearRGB(b);
	}

	a *= validatedExtraAlpha;

	return ((a << 24) | (r << 16) | (g << 8) | (b));
    }

    /**
     * The currently only known paint XRender can't support are radial gradients
     * where focus and center paint differ.
     */
    public static boolean isPaintValid(SunGraphics2D sg2d) {
	if (sg2d.paintState == SunGraphics2D.PAINT_RAD_GRADIENT) {
	    RadialGradientPaint grad = (RadialGradientPaint) sg2d.paint;
	    return grad.getFocusPoint().equals(grad.getCenterPoint());
	}
	return true;
    }

    /**
     * Sets a 2-stop gradient. Utility method generating the by the native
     * method expected parameters and calling it.
     */
    private void setGradientPaint(SunGraphics2D sg2d, GradientPaint paint, boolean useMask) {
	int[] pixels = convertToIntArgbPixels(new Color[] { paint.getColor1(), paint.getColor2() }, false);

	float fractions[] = new float[2];
	fractions[0] = 0;
	fractions[1] = 1;

	Point2D pt1 = paint.getPoint1();
	Point2D pt2 = paint.getPoint2();

	AffineTransform at = (AffineTransform) sg2d.transform.clone();
	try {
	    at.invert();
	} catch (NoninvertibleTransformException ex) {
	    ex.printStackTrace();
	}

	double[] transformData = new double[6];
	at.getMatrix(transformData);

	int repeat = paint.isCyclic() ? XRUtils.RepeatReflect : XRUtils.RepeatPad;

	XRSetLinearGradientPaint(fractions, pixels, XDoubleToFixed(pt1.getX()), XDoubleToFixed(pt1.getY()), XDoubleToFixed(pt2.getX()),
		XDoubleToFixed(pt2.getY()), 2, repeat, useMask, false, XDoubleToFixed(transformData[0]), XDoubleToFixed(transformData[2]),
		XDoubleToFixed(transformData[4]), XDoubleToFixed(transformData[1]), XDoubleToFixed(transformData[3]),
		XDoubleToFixed(transformData[5]));
    }

    /**
     * Sets a n-stop linear gradient. Utility method generating the by the
     * native method expected parameters and calling it.
     */
    public void setLinearGradientPaint(SunGraphics2D sg2d, LinearGradientPaint paint, boolean useMask) {
	boolean linear = (paint.getColorSpace() == ColorSpaceType.LINEAR_RGB);
	Color[] colors = paint.getColors();
	int numStops = colors.length;
	Point2D pt1 = paint.getStartPoint();
	Point2D pt2 = paint.getEndPoint();

	AffineTransform at = paint.getTransform();
	at.preConcatenate(sg2d.transform);

	int cycleMethod = XRUtils.getRepeatForCycleMethod(paint.getCycleMethod());
	float[] fractions = paint.getFractions();
	int[] pixels = convertToIntArgbPixels(colors, linear);

	try {
	    at.invert();
	} catch (NoninvertibleTransformException ex) {
	    ex.printStackTrace();
	}

	double[] transformData = new double[6];
	at.getMatrix(transformData);

	XRSetLinearGradientPaint(fractions, pixels, XDoubleToFixed(pt1.getX()), XDoubleToFixed(pt1.getY()), XDoubleToFixed(pt2.getX()),
		XDoubleToFixed(pt2.getY()), numStops, cycleMethod, useMask, linear, XDoubleToFixed(transformData[0]),
		XDoubleToFixed(transformData[2]), XDoubleToFixed(transformData[4]), XDoubleToFixed(transformData[1]),
		XDoubleToFixed(transformData[3]), XDoubleToFixed(transformData[5]));
    }

    /**
     * Sets a Radial Gradient where focus and center point match. Utility method
     * generating the by the native method expected parameters and calling it.
     */
    public void setRadialGradientPaint(SunGraphics2D sg2d, RadialGradientPaint paint) {
	boolean linear = (paint.getColorSpace() == ColorSpaceType.LINEAR_RGB);
	Color[] colors = paint.getColors();
	int numStops = colors.length;
	Point2D center = paint.getCenterPoint();
	Point2D focus = paint.getFocusPoint();

	int cycleMethod = XRUtils.getRepeatForCycleMethod(paint.getCycleMethod());
	float[] fractions = paint.getFractions();
	int[] pixels = convertToIntArgbPixels(colors, linear);
	float radius = paint.getRadius();

	// save original (untransformed) center and focus points
	double cx = center.getX();
	double cy = center.getY();
	double fx = focus.getX();
	double fy = focus.getY();

	AffineTransform at = paint.getTransform();
	at.preConcatenate(sg2d.transform);
	focus = at.transform(focus, focus);

	// transform unit circle to gradient coords; we start with the
	// unit circle (center=(0,0), focus on positive x-axis, radius=1)
	// and then transform into gradient space
	at.translate(cx, cy);
	at.rotate(fx - cx, fy - cy);
	// at.scale(radius, radius);

	// invert to get mapping from device coords to unit circle
	try {
	    at.invert();
	} catch (Exception e) {
	    at.setToScale(0.0, 0.0);
	}
	focus = at.transform(focus, focus);

	// clamp the focus point so that it does not rest on, or outside
	// of, the circumference of the gradient circle
	fx = Math.min(focus.getX(), 0.99);

	double[] transformData = new double[6];
	at.getMatrix(transformData);

	XRSetRadialGradientPaint(fractions, pixels, XDoubleToFixed(fx), numStops, cycleMethod, false, linear, XDoubleToFixed(0),
		XDoubleToFixed(radius), XDoubleToFixed(transformData[0]), XDoubleToFixed(transformData[2]), XDoubleToFixed(transformData[4]),
		XDoubleToFixed(transformData[1]), XDoubleToFixed(transformData[3]), XDoubleToFixed(transformData[5]));
    }

    private TexturePaint setTexturePaint(SunGraphics2D sg2d, TexturePaint paint, boolean useMask) {
	BufferedImage bi = paint.getImage();
	SurfaceData dstData = sg2d.surfaceData;
	SurfaceData srcData = dstData.getSourceSurfaceData(bi, SunGraphics2D.TRANSFORM_ISIDENT, CompositeType.SrcOver, null);

	// REMIND: this hack tries to ensure that we have a cached texture
	if (!(srcData instanceof XRSurfaceData)) {
	    srcData = dstData.getSourceSurfaceData(paint.getImage(), SunGraphics2D.TRANSFORM_ISIDENT, CompositeType.SrcOver, null);
	    if (!(srcData instanceof XRSurfaceData)) {
		return null;
	    }
	}

	XRSurfaceData x11SrcData = (XRSurfaceData) srcData;

	AffineTransform at = (AffineTransform) sg2d.transform.clone();
	Rectangle2D anchor = paint.getAnchorRect();
	at.translate(anchor.getX(), anchor.getY());
	at.scale(anchor.getWidth() / ((double) bi.getWidth()), anchor.getHeight() / ((double) bi.getHeight()));

	try {
	    at.invert();
	} catch (NoninvertibleTransformException ex) {
	    at.setToIdentity(); /* TODO: Right thing to do in this case? */
	}
	
	x11SrcData.validateAsSource(at, XRUtils.RepeatNormal, XRUtils.ATransOpToXRQuality(sg2d.interpolationType));

	XRSetTexturePaint(srcData.getNativeOps());

	return paint;
    }

    /**
     * Sets the paint as source. Checks the type of paint and calls the
     * appropriate set*Paint method.
     * 
     * @param sg2d
     * @param paint
     */
    public void setPaint(SunGraphics2D sg2d, Paint paint) {
	if (sg2d != null) {
	    if (sg2d.paintState <= SunGraphics2D.PAINT_ALPHACOLOR) {
		XRResetPaint();
	    } else {
		switch (sg2d.paintState) {
		case SunGraphics2D.PAINT_GRADIENT:
		    setGradientPaint(sg2d, (GradientPaint) paint, false);
		    validatedPaint = paint;
		    break;

		case SunGraphics2D.PAINT_LIN_GRADIENT:
		    setLinearGradientPaint(sg2d, (LinearGradientPaint) paint, false);
		    validatedPaint = paint;
		    break;

		case SunGraphics2D.PAINT_RAD_GRADIENT:
		    setRadialGradientPaint(sg2d, (RadialGradientPaint) paint);
		    validatedPaint = paint;
		    break;

		case SunGraphics2D.PAINT_TEXTURE:
		    validatedPaint = setTexturePaint(sg2d, (TexturePaint) paint, false);
		    break;

		default:
		    throw new InternalError("should not reach");
		}
	    }
	}
    }

    /**
     * Validates the Surface when used as destination, takes care that the
     * native surface has the same state as expected, changing it it not.
     */
    public void validate(Region clip, Composite comp, AffineTransform xform, Paint paint, SunGraphics2D sg2d, int flags) {
	boolean updateClip = (clip != validatedClip);
	boolean updatePaint = (paint != validatedPaint) || paint == null;

	if (!isValid()) {
	    throw new InvalidPipeException("bounds changed");
	}

	// validate clip
	if (updateClip) {
	    if (clip != null) {
		//    System.out.println("Set paint clip: "+getBounds().width+":"+getBounds().height);
		//    Thread.dumpStack();
		XRSetClip(getNativeOps(), clip.getLoX(), clip.getLoY(), clip.getHiX(), clip.getHiY(), clip.isRectangular() ? null : clip);
	    } else {
		XRResetClip(getNativeOps());
	    }
	    validatedClip = clip;
	}

	// validate composite (note that a change in the context flags
	// may require us to update the composite state, even if the
	// composite has not changed)
	if ((comp != validatedComp) || (flags != validatedFlags)) {
	    if (comp != null) {
		setComposite(comp);
	    } else {
		comp = AlphaComposite.getInstance(AlphaComposite.SRC_OVER);
		setComposite(comp);
	    }
	    // the paint state is dependent on the composite state, so make
	    // sure we update the color below
	    updatePaint = true;
	    validatedComp = comp;
	    validatedFlags = flags;
	}

	if (sg2d != null && validatedPixel != sg2d.pixel) {
	    validatedPixel = sg2d.pixel;
	    XRSetForeground(validatedPixel);
	}

	// validate paint
	if (updatePaint) {
	    if (paint != null) {
		setPaint(sg2d, paint);
	    } else {
		XRResetPaint();
	    }
	    validatedPaint = paint;
	}
    }

    public void makePipes() { /*
				 * TODO: Why was this synchronized, but access
				 * not?
				 */
	if (xrpipe == null) {
	    try {
		SunToolkit.awtLock();
		xgc = XCreateGC(getNativeOps());

		xrpipe = XRRenderer.getInstance();
		xrtxpipe = new PixelToShapeConverter(xrpipe);
		xrtextpipe = new XRTextRenderer();
		xrDrawImage = new XRDrawImage();
	    } finally {
		SunToolkit.awtUnlock();
	    }
	}
    }

    public static class XRWindowSurfaceData extends XRSurfaceData {
	public XRWindowSurfaceData(X11ComponentPeer peer, XRGraphicsConfig gc, SurfaceType sType) {
	    super(peer, gc, sType, peer.getColorModel(), peer.getColorModel().getPixelSize(), Transparency.OPAQUE);

	    if (isXRDrawableValid()) {
		initXRender(XRUtils.getPictureFormatForTransparency(Transparency.OPAQUE));
		makePipes();
	    }
	}

	public SurfaceData getReplacement() {
	    return peer.getSurfaceData();
	}

	public Rectangle getBounds() {
	    Rectangle r = peer.getBounds();
	    r.x = r.y = 0;
	    return r;
	}

	@Override
	public boolean canSourceSendExposures(int x, int y, int w, int h) {
	    return true;
	}

	/**
	 * Returns destination Component associated with this SurfaceData.
	 */
	public Object getDestination() {
	    return peer.getTarget();
	}
    }

    public static class XRPixmapSurfaceData extends XRSurfaceData {
	Image offscreenImage;
	int width;
	int height;
	int transparency;

	public XRPixmapSurfaceData(XRGraphicsConfig gc, int width, int height, Image image, SurfaceType sType, ColorModel cm, long drawable,
		int transparency, int pictFormat, int depth) {
	    super(null, gc, sType, cm, depth, transparency);
	    this.width = width;
	    this.height = height;
	    offscreenImage = image;
	    this.transparency = transparency;
	    initSurface(depth, width, height, drawable, pictFormat);
	    makePipes();
	}

	public void initSurface(int depth, int width, int height, long drawable, int pictFormat) {
	    try {
		SunToolkit.awtLock();
		XRInitSurface(depth, width, height, drawable, pictFormat);
	    } finally {
		SunToolkit.awtUnlock();
	    }
	}

	public SurfaceData getReplacement() {
	    return restoreContents(offscreenImage);
	}

	/**
	 * Need this since the surface data is created with the color model of
	 * the target GC, which is always opaque. But in SunGraphics2D.blitSD we
	 * choose loops based on the transparency on the source SD, so it could
	 * choose wrong loop (blit instead of blitbg, for example).
	 */
	public int getTransparency() {
	    return transparency;
	}

	public Rectangle getBounds() {
	    return new Rectangle(width, height);
	}

	@Override
	public boolean canSourceSendExposures(int x, int y, int w, int h) {
	    return (x < 0 || y < 0 || (x + w) > width || (y + h) > height);
	}

	public void flush() {
	    /*
	     * We need to invalidate the surface before disposing the native
	     * Drawable and Picture. This way if an application tries to render
	     * to an already flushed XRSurfaceData, we will notice in the
	     * validate() method above that it has been invalidated, and we will
	     * avoid using those native resources that have already been
	     * disposed.
	     */
	    invalidate();
	    flushNativeSurface();
	}

	/**
	 * Returns destination Image associated with this SurfaceData.
	 */
	public Object getDestination() {
	    return offscreenImage;
	}
    }

    private static LazyPipe lazypipe = new LazyPipe();

    public static class LazyPipe extends ValidatePipe {
	public boolean validate(SunGraphics2D sg2d) {
	    XRSurfaceData xsd = (XRSurfaceData) sg2d.surfaceData;
	    if (!xsd.isXRDrawableValid()) {
		return false;
	    }
	    xsd.makePipes();
	    return super.validate(sg2d);
	}
    }
}
